/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/
#define FILECODE "F(301)"

/*
 * vmx86.c --
 *
 *     Platform independent routines for creating/destroying/running
 *     virtual machine monitors.
 *
 */

#ifdef linux
/* Must come before any kernel header file --hpreg */
#   include "driver-config.h"

#   include <linux/string.h> /* memset() in the kernel */
#   include <linux/sched.h> /* jiffies from the kernel */
#elif defined(WINNT_DDK)
#   include <string.h>
#else
#   error "Unknown platform"
#endif

#include "machine.h"
#include "vm_types.h"
#include "vm_assert.h"
#include "modulecall.h"
#include "vmx86.h"
#include "task.h"
#include "initblock.h"
#include "vm_asm.h"
#include "vtrace.h"
#include "memmgmt.h"
#include "wslimits.h"
#ifdef USE_PERFCTRS_HOSTED
#   include "perfctr.h"
#endif
#include "hostif.h"


/*
 * Keep track of the virtual machines that have been
 * created using the following structures.
 *
 */

static VMDriver *vmDriverList = NULL;

static Bool  lockedPageLimitMonitor;
static uintptr_t lockedPageLimit;
static uintptr_t numLockedPages;
static uintptr_t vmCount;

static void FreeAllResources(VMDriver *vm);

static uintptr_t unlockedPages;
static uintptr_t dirtyUnlockedPages;


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_CreateVM --
 *
 *      Allocate and initialize a driver structure for a virtual machine.
 *
 * Results:
 *      VMDriver structure or NULL on error.
 *
 * Side effects:
 *       Memory allocated.
 *
 *----------------------------------------------------------------------
 */

VMDriver *
Vmx86_CreateVM(void *uniqHandle,             // IN: Unique id for VM to be
                                             //     created
	       uintptr_t processId,          // IN: Process creating VM
	       WslimitsInfo *wslimitsInfoIn) // IN: Information about resource
                                             //     limits
{
   VMDriver *vm;
   VMDriver **vmp;
   
   /*
    * For NT:
    * The code within a critical region guarded by an spin lock must
    * neither be pageable nor make any references to pageable data.
    */

   vm = HostIF_AllocKernelMem(sizeof *vm, TRUE);
   if (vm == NULL) {
      return NULL;
   }
   memset(vm, 0, sizeof *vm);
#ifdef linux
   vm->logFD = -1;
#endif
   HostIF_GlobalVMLock(0);

   if (vmCount >= MAX_VMS) {
      HostIF_GlobalVMUnLock(0);
      HostIF_FreeKernelMem(vm);
      return NULL;
   }
   vmCount++;

   vm->count++;
   for (vmp = &vmDriverList; *vmp != NULL; vmp = &(*vmp)->nextDriver) {
   }
   *vmp = vm;

   ASSERT(sizeof (uintptr_t) == sizeof vm);
   vm->id = ((uintptr_t)vm & 0xffffffff) >> 1;
   vm->uniqHandle = uniqHandle;
   vm->processID = processId;

   memcpy(&vm->wslimitsInfo, (char *)wslimitsInfoIn, sizeof vm->wslimitsInfo);

   HostIF_GlobalVMUnLock(0);

   if (HostIF_Init(vm)) {
      Vmx86_ReleaseVM(vm);
      return NULL;
   }

   return vm;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_BindVM  --
 *
 *      Bind to an existing VM.
 *
 * Results:
 *      VMDriver structure or NULL on error.
 *
 * Side effects:
 *	Increment VM reference count.
 *
 *----------------------------------------------------------------------
 */

VMDriver *
Vmx86_BindVM(int id)
{
   VMDriver *vm;

   HostIF_GlobalVMLock(14);
   for (vm = vmDriverList; vm != NULL; vm = vm->nextDriver) {
      if (vm->id == id) {
	 vm->count++;
	 break;
      }
   }
   HostIF_GlobalVMUnLock(14);
   return vm;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_ReleaseVM  --
 *
 *      Release a VM (either created here or from a bind).
 *
 * Results:
 *      zero if successful
 *
 * Side effects:
 *	Decrement VM reference count.
 *      Release resources (those that are left) when count reaches 0.
 *
 *----------------------------------------------------------------------
 */
int
Vmx86_ReleaseVM(VMDriver *vm)
{
   VMDriver **vmp;

   HostIF_GlobalVMLock(1);

   /*
    * Do reference counting first
    */

   if (--vm->count > 0) {
      HostIF_GlobalVMUnLock(1);
      return 0;
   }

   ASSERT(vm->count == 0);

   /*
    *   Deletion of the virtual machine from list
    */

   for (vmp = &vmDriverList; *vmp != vm; vmp = &(*vmp)->nextDriver) {
      ASSERT(*vmp != NULL);
   }
   *vmp = vm->nextDriver;

   vmCount--;

   HostIF_GlobalVMUnLock(1);

   Log("Vmx86_ReleaseVM: unlocked pages: %d, unlocked dirty pages: %d\n",
       unlockedPages, dirtyUnlockedPages);

   FreeAllResources(vm);

   return 0;
}

/*
 *------------------------------------------------------------------------------
 *
 * Vmx86_InitVM --
 *
 *    Initializaiton of the VM.  Expects all initial arguments
 *    to be part of the InitBlock structure.
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    Many
 *
 *------------------------------------------------------------------------------
 */

int
Vmx86_InitVM(VMDriver *vm,          // IN
             InitBlock *initParams) // IN: Initial params from the VM
{
   int retval;
   Vcpuid vcpuid;

   if (initParams->magicNumber != INIT_BLOCK_MAGIC) {
      Warning("Bad magic number for init block 0x%x\n", initParams->magicNumber);
      return 1;
   }
   if (initParams->numVCPUs >= MAX_INITBLOCK_CPUS) {
      Warning("Too many VCPUs for init block %d\n", initParams->numVCPUs);
      return 1;
   }
   vm->numVCPUs = initParams->numVCPUs;
   vm->logFD    = initParams->logFD;

   HostIF_InitFP(vm);
   HostIF_InitEvent(vm);

   /*
    * Initialize the driver's part of the cross-over page used to
    * talk to the monitor
    */

   retval = Task_InitCrosspage(vm,initParams);
   if (retval) {
      Warning("Task crosspage init died with retval=%d\n", retval);
      goto error;
   }
   for (vcpuid = 0; vcpuid < vm->numVCPUs;  vcpuid++) {
      vm->crosspage[vcpuid]->userCallRequest = MODULECALL_USERCALL_NONE;
      vm->crosspage[vcpuid]->moduleCallInterrupted = FALSE;
   }

   unlockedPages = dirtyUnlockedPages = 0;
   return 0;

error:
   FreeAllResources(vm);
   return 1;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_LateInitVM --
 *
 *      Do late initialization of the driver.
 *	This should be called after Vmx86_CreateVM and
 *	after all the user-level device initialization.
 *
 * Results: 
 *	non-zero on error, zero on success;
 *
 * Side effects:
 *      VTrace_Init is called
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_LateInitVM(VMDriver *vm)
{
   VTrace_Init(vm);
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_RunVM  --
 *
 *      Main interaction between the module and the monitor:
 *
 *	   Run the monitor
 *	   Process module calls from the monitor
 *	   Make cross user calls to the main thread
 *	   Return to userlevel to process normal user calls
 *	   and to signal timeout or errors.
 *
 * Results:
 *      Positive: user call number.
 *	-1: error (currently not used).
 *
 * Side effects:
 *      Not really.
 *
 *----------------------------------------------------------------------
 */

int
Vmx86_RunVM(VMDriver *vm, Vcpuid vcpuid)
{
   uint32 retval = MODULECALL_USERRETURN;
   VMCrossPage *crosspage = vm->crosspage[vcpuid];

   ASSERT(crosspage);

   /*
    * Check if we were interrupted by signal.
    */
   if (crosspage->moduleCallInterrupted) {
      crosspage->moduleCallInterrupted = FALSE;
      goto skipTaskSwitch;
   }
   for (;;) {

      crosspage->retval = retval;

      /*
       * Task_Switch changes the world to the monitor.
       * The monitor is waiting in the BackToHost routine.
       */
      UCTIMESTAMP(crosspage, SWITCHING_TO_MONITOR);
      Task_Switch(vm, vcpuid);
      UCTIMESTAMP(crosspage, SWITCHED_TO_MODULE);

skipTaskSwitch:;

      retval = MODULECALL_USERRETURN;

      if (crosspage->userCallType != MODULECALL_USERCALL_NONE) {
         /*
          * This is the main user call path.
          *
          * There are two kinds of user calls.  Normal ones
          * are handled by the calling VCPU thread itself.
          * We just return from here (back to userlevel)
          * in that case.
          *
          * Calls marked userCallCross are handled by the main
          * VMX thread.  In this case, the userCallRequest field
          * indicates to the VMX that this VCPU wants to make
          * a user call.  This field may be consulted by the VMX
          * at any time (specifically when the VMX is awakened by
          * another VCPU), so it must be set after the other
          * user call arguments.  The VMX is responsible for
          * resetting this field and awakening the VCPU when
          * the user call is complete, via the ACK_USER_CALL
          * and COMPLETE_USER_CALL ioctl.  The latter implies
          * the former.
          *
          * When and how to use ACK_USER_CALL and COMPLETE_USER_CALL
          * are at the discretion of the VMX.  In particular,
          * COMPLETE_USER_CALL does not imply that the requested
          * operation has fully completed, only that the VCPU can
          * continue.  See the comment in MonitorLoopHandleUserCall()
          * for use scenarios.
          *
          * See also comment at MonitorLoopUserCallPoll().
          *
          * -- edward
          */

         if (!crosspage->userCallCross) {
            ASSERT(!crosspage->userCallRestart);
            return crosspage->userCallType;
         }

	 if (!crosspage->userCallRestart) {
	    ASSERT(crosspage->userCallRequest == MODULECALL_USERCALL_NONE);
	    crosspage->userCallRequest = crosspage->userCallType;
	    UCTIMESTAMP(crosspage, AWAKENING_VMX);
	    HostIF_UserCall(vm, vcpuid);
	 }

	 UCTIMESTAMP(crosspage, GOING_TO_SLEEP);
         if (HostIF_UserCallWait(vm, vcpuid)) {
	    ASSERT(crosspage->userCallRequest == MODULECALL_USERCALL_NONE);
	 } else {
	    retval = MODULECALL_USERTIMEOUT;
	 }
	 UCTIMESTAMP(crosspage, AWAKE);
      }

      switch (crosspage->moduleCallType) {

      case MODULECALL_NONE:
         break;

      case MODULECALL_INTR:    // Already done in task.c
         break;

      case MODULECALL_ISMPNLOCKED : {
         MPN mpn = crosspage->args[0];
         retval = HostIF_IsMPNLocked(vm, mpn);
         break;
      }

      case MODULECALL_LOCKPAGE: {
         void *addr = (void *)crosspage->args[0];
#ifdef linux
	 HostIF_LockKernel();
#endif
	 ASSERT(sizeof retval == sizeof (MPN));
         retval = Vmx86_LockPage(vm, addr, FALSE);
	 if (PAGE_LOCK_SUCCESS(retval)) {
            if (HostIF_MarkPageDirty(addr)) {
               retval = Vmx86_UnlockPage(vm, addr, FALSE);
               if (PAGE_LOCK_SUCCESS(retval)) {
                  retval = PAGE_LOCK_TOUCH_FAILED;
               }
            }
         }
#ifdef linux
	 HostIF_UnLockKernel();
#endif
         break;
      }

      case MODULECALL_GET_RECYCLED_PAGE:
#ifdef linux
	 HostIF_LockKernel();
#endif
	 retval = Vmx86_GetRecycledPage(vm);
#ifdef linux
	 HostIF_UnLockKernel();
#endif
	 break;

      case MODULECALL_SEMAWAIT:
         retval = HostIF_SemaphoreWait(vm, vcpuid, crosspage->args[0]);
	 if (retval == MX_WAITINTERRUPTED) {
	    crosspage->moduleCallInterrupted = TRUE;
	    return USERCALL_RESTART;
	 }
         break;
      case MODULECALL_SEMASIGNAL:
         retval = HostIF_SemaphoreSignal(crosspage->args[0]);
	 if (retval == MX_WAITINTERRUPTED) {
	    crosspage->moduleCallInterrupted = TRUE;
	    return USERCALL_RESTART;
	 }
         break;
      case MODULECALL_SEMAFORCEWAKEUP: {
         HostIF_SemaphoreForceWakeup(vm, (Vcpuid) crosspage->args[0]);
         break;
      }
      case MODULECALL_IPI:
         retval = HostIF_IPI((HostThread)crosspage->args[0]);
         break;

      case MODULECALL_UNLOCKPAGE: {
         void *addr = (void *)crosspage->args[0];
         Bool dirty = (Bool)crosspage->args[1];

#ifdef linux
	 HostIF_LockKernel();
#endif
         unlockedPages++;
         if (dirty) {
            dirtyUnlockedPages++;
	    if (HostIF_MarkPageDirty(addr)) {
	       retval = PAGE_UNLOCK_TOUCH_FAILED;
	    } else {
	       retval = Vmx86_UnlockPage(vm, addr, FALSE);
	    }
         } else {
	    retval = Vmx86_UnlockPage(vm, addr, FALSE);
	 }
#ifdef linux
	 HostIF_UnLockKernel();
#endif
         break;
      }

      case MODULECALL_RELEASE_ANON_PAGE: {
         MPN mpn = (MPN)crosspage->args[0];
#ifdef linux
	 HostIF_LockKernel();
#endif
	 retval = Vmx86_ReleaseAnonPage(vm, mpn);
#ifdef linux
	 HostIF_UnLockKernel();
#endif
         break;
      }

      default:
         Warning("ModuleCall %d not supported\n", crosspage->moduleCallType);
      }

      /*
       * Give other threads and processes a chance.
       *
       * This is important when we are not preemptable while
       * in the driver and the monitor (e.g., Linux).
       */

      HostIF_Reschedule();
   }

   NOT_REACHED();
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetStartTime --
 *
 *      Initial the data structures that track the starting time of the
 *      virtual machine.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      none
 *
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_SetStartTime(VmTimeStart *st)	// OUT: return value
{
   uint32 flags;

   SAVE_FLAGS(flags);
   CLEAR_INTERRUPTS();

   st->count = GET_TSC();
   st->time = HostIF_ReadUptime();

   RESTORE_FLAGS(flags);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetkHzEstimate
 *
 *      Return an estimate the of the processor's kHz rating, based on
 *      the ratio of the cycle counter and system uptime since the
 *      driver was loaded.
 *      This function could be called (on Windows) at IRQL DISPATCH_LEVEL.
 *
 * Results:
 *      Processor speed in kHz.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

uint32
Vmx86_GetkHzEstimate(VmTimeStart *st)	// IN: start time
{
   uint64 cDiff, tDiff, freq, hz;
   uint32 tmp;
   static uint32 kHz; 
#ifdef linux
   uint32 tmpkHz;
#endif

   /* 
    * Cache and return the first result for consistency. 
    * TSC values can be changed without notification.
    * TSC frequency can be vary too (SpeedStep, slowing clock on HALT, etc.)
    */
   if (kHz != 0) {
      return kHz;
   }

   SAVE_FLAGS(tmp);
   CLEAR_INTERRUPTS();

   cDiff = GET_TSC() - st->count;
   tDiff = HostIF_ReadUptime() - st->time;

   RESTORE_FLAGS(tmp);

   if (tDiff == 0) {
      return 0;  /* Failure */
   }

   /*
    * Compute the CPU speed in kHz, which is cDiff / (tDiff /
    * HostIF_UptimeFrequency()) / 1000.  We need to do the computation
    * carefully to avoid overflow or undue loss of precision.  Also,
    * on Linux we can't do a 64/64=64 bit division directly, as the
    * gcc stub for that is not linked into the kernel.
    */
   freq = HostIF_UptimeFrequency();
#ifndef linux
   while (cDiff > ((uint64) -1) / freq) {
      cDiff >>= 1;
      tDiff >>= 1;
   }
   hz = (cDiff * freq) / tDiff;
   kHz = (hz + 500) / 1000;
#else
   /* On Linux we can't do a 64/64=64 bit division, as the gcc stub
    * for that is not linked into the kernel.  We'll assume that cDiff
    * * freq fits into 64 bits and that tDiff fits into 32 bits.  This
    * is safe given the values used on Linux.
    */
   Div643264(cDiff * freq, tDiff, &hz, &tmp);
   Div643232(hz + 500, 1000, &tmpkHz, &tmp);
   kHz = tmpkHz;
#endif
   return kHz;

}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_CurrentVM --
 *
 *      Return the VMDriver structure associated with
 *      the handle
 *
 *
 * Results:
 *      VMDriver * structure.NULL if not available
 *
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
VMDriver *
Vmx86_CurrentVM(void *uniqHandle)
{
  VMDriver *vm;

  HostIF_GlobalVMLock(2);

  vm = vmDriverList;

  while (vm) {
     if (vm->uniqHandle == uniqHandle) {
        break;
     }
     vm = vm->nextDriver;
  }

  HostIF_GlobalVMUnLock(2);

  return vm;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetVMForProcess  --
 *
 *      Return the VMDriver structure associated with
 *      a process
 *
 *
 * Results:
 *      VMDriver * structure.NULL if not available
 *
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

VMDriver *
Vmx86_GetVMforProcess(uintptr_t processId)
{
  VMDriver *vm;

  HostIF_GlobalVMLock(3);

  vm = vmDriverList;

  while (vm) {
    if (vm->processID == processId) {
       break;
    }
    vm = vm->nextDriver;
  }

  HostIF_GlobalVMUnLock(3);

  return vm;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetNumVMs  --
 *
 *      Return the number of VMs.
 *
 * Results:
 *      The number of VMs.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
int32
Vmx86_GetNumVMs()
{
   return vmCount;
}

int32
Vmx86_GetTotalMemUsage()
{
   VMDriver *vm;
   int totalmem = 0;

   HostIF_GlobalVMLock(15);
   vm = vmDriverList;

   while (vm) {
      totalmem += vm->wslimitsInfo.memsize;
      vm = vm->nextDriver;
   }
   
   HostIF_GlobalVMUnLock(15);
   return totalmem;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetLockedPagesLimit --
 *
 *      Set the hard limit for the number of locked pages.
 *      The limit can be set only when one VM is running or if it not set.
 *
 * Results:
 *      Returns TRUE on success and FALSE on failure to set the limit
 *
 * Side effects:
 *      Hard limit changed.
 *
 *----------------------------------------------------------------------
 */
Bool
Vmx86_SetLockedPagesLimit(uintptr_t limit)
{
   Bool retval = FALSE;

   HostIF_GlobalVMLock(4);
   if (!lockedPageLimitMonitor && 
		 (lockedPageLimit == 0 || (lockedPageLimit != 0 && vmCount == 1))) {
      lockedPageLimit = limit;
      retval = TRUE;
   }
   HostIF_GlobalVMUnLock(4);

   return retval;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetLockedPageLimitMonitorStatus --
 *
 *      Set the status of the hard limit monitor - 1 is on, 0 is off.
 *
 * Results:
 *      Returns TRUE on success and FALSE on failure to set the limit
 *
 * Side effects:
 *      Hard limit changed.
 *
 *----------------------------------------------------------------------
 */
void
Vmx86_SetLockedPageLimitMonitorStatus(Bool status)
{
   lockedPageLimitMonitor = status;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_ChangeLockedPagesLimit --
 *
 *      Change the hard limit for the number of locked pages.
 *      The limit can be reduced at any time.
 *
 * Results:
 *      Returns new limit.
 *
 * Side effects:
 *      Hard limit changed.
 *
 *----------------------------------------------------------------------
 */
uintptr_t
Vmx86_ChangeLockedPagesLimit(uintptr_t limit)
{
   HostIF_GlobalVMLock(11);
   lockedPageLimit = limit;
   HostIF_GlobalVMUnLock(11);

   return lockedPageLimit;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetLockedPagesLimit --
 *
 *      Get the hard limit for the number of locked pages.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Hard limit changed.
 *
 *----------------------------------------------------------------------
 */
int32
Vmx86_GetLockedPagesLimit()
{
   return lockedPageLimit;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_LockPage --
 *
 *      Lock a page.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Number of global and per-VM locked pages increased.
 *
 *----------------------------------------------------------------------
 */
MPN
Vmx86_LockPage(VMDriver *vm, void * addr, Bool userPage)
{
   MPN mpn;

   if (lockedPageLimit != 0 && numLockedPages >= lockedPageLimit) {
      return PAGE_LOCK_LIMIT_EXCEEDED;
   }

   mpn = HostIF_LockPage(vm,addr);
   if (!PAGE_LOCK_SUCCESS(mpn)) {
      return mpn;
   }

   HostIF_GlobalVMLock(5);
   numLockedPages++;
   HostIF_GlobalVMUnLock(5);

   vm->memMgmtInfo.currentLocked++;
   if (userPage) {
      vm->stats.userLockedPages++;
      if (vm->stats.userLockedPages > vm->stats.maxUserLockedPages) {
	 vm->stats.maxUserLockedPages = vm->stats.userLockedPages;
      }
   } else {
      vm->stats.monitorLockedPages++;
      if (vm->stats.monitorLockedPages > vm->stats.maxMonitorLockedPages) {
	 vm->stats.maxMonitorLockedPages = vm->stats.monitorLockedPages;
      }
   }
   return mpn;
}



/*
 *----------------------------------------------------------------------
 *
 * Vmx86_UnlockPage --
 *
 *      Unlock a page.
 *
 * Results:
 *      
 *
 * Side effects:
 *      Number of global and per-VM locked pages decreased.
 *
 *----------------------------------------------------------------------
 */
int
Vmx86_UnlockPage(VMDriver *vm, void *addr, Bool userPage)
{
   int retval = HostIF_UnlockPage(vm, addr);

   if (PAGE_LOCK_SUCCESS(retval)) {
      HostIF_GlobalVMLock(6);
      numLockedPages--;
      HostIF_GlobalVMUnLock(6);

      vm->memMgmtInfo.currentLocked--;
      if (userPage) {
         vm->stats.userLockedPages--;
      } else {
         vm->stats.monitorLockedPages--;
      }
   }
   return retval;
}



/*
 *----------------------------------------------------------------------
 *
 * Vmx86_ReleaseAnonPage --
 *
 *      Unlock a page.
 *
 * Results:
 *      PAGE_UNLOCK_NO_ERROR
 *
 * Side effects:
 *      Number of global and per-VM locked pages decreased.
 *
 *----------------------------------------------------------------------
 */
int
Vmx86_ReleaseAnonPage(VMDriver *vm, MPN mpn)
{
   int retval;

   retval = HostIF_ReleaseAnonPage(vm, mpn);
   if (PAGE_LOCK_SUCCESS(retval)) {
      HostIF_GlobalVMLock(20);
      numLockedPages--;
      vm->memMgmtInfo.currentLocked--;
      vm->stats.monitorLockedPages--;
      HostIF_GlobalVMUnLock(20);
   }
   return retval;
}



MPN
Vmx86_GetRecycledPage(VMDriver *vm)
{
   VPN vpn;

   vpn = HostIF_GetRecycledPage(vm);
   if (vpn != INVALID_VPN) {
      MPN mpn = Vmx86_LockPage(vm, (void*)VPN_2_VA(vpn), FALSE);
      if (PAGE_LOCK_SUCCESS(mpn)) {
	 return mpn;
      }
   }
   return INVALID_MPN;
}



/*
 *----------------------------------------------------------------------
 *
 * Vmx86_GetMemInfo --
 *
 *      Return the info about all VMs.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      VMGetMemInfoArgs is filled in.
 *
 *----------------------------------------------------------------------
 */
Bool
Vmx86_GetMemInfo(VMDriver *curVM,
                 int32 curVMOnly,
                 VMGetMemInfoArgs *outArgs,
                 int outArgsLength)
{
   VMDriver *vm;
   int i;
   int outSize;
   int wantedVMs;

   HostIF_GlobalVMLock(7);

   if (curVMOnly) {
      wantedVMs = 1;
   } else {
      wantedVMs = vmCount;
   }

   outSize = VM_GET_MEM_INFO_SIZE(wantedVMs);
   if (outSize > outArgsLength) {
      HostIF_GlobalVMUnLock(7);
      return FALSE;
   }

   outArgs->numVMs = wantedVMs;
   outArgs->numLockedPages = numLockedPages;
   outArgs->maxLockedPages = lockedPageLimit;
   outArgs->callerIndex = -1;

   if (curVM != NULL) {
      if (wantedVMs == 1) {
         outArgs->memMgmtInfo[0] = curVM->memMgmtInfo;
         outArgs->callerIndex = 0;
      } else {
         vm = vmDriverList;
         i = 0;
         outArgs->callerIndex = -1;
         while (vm != NULL && i < vmCount) {
	    if (vm == curVM) {
	       outArgs->callerIndex = i;
	    }
	    outArgs->memMgmtInfo[i] = vm->memMgmtInfo;
	    i++;

	    vm = vm->nextDriver;
         }
      }
   }

   HostIF_GlobalVMUnLock(7);

   return TRUE;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Vmx86_GetMemInfoCopy --
 *
 *    Return the information about all VMs by copying the data out to user
 *    memory.
 *
 *    On input, outArgs->numVMs indicates how much space has been allocated for
 *    the information. On output, it indicates how much space has been
 *    filled --hpreg
 *
 * Results:
 *    TRUE on success
 *    FALSE on failure
 *
 * Side effects:
 *    None
 *
 *-----------------------------------------------------------------------------
 */

Bool
Vmx86_GetMemInfoCopy(VMDriver *curVM,           // IN
                     VMGetMemInfoArgs *outArgs) // IN/OUT
{
   /*
    * The union guarantees a correct alignment
    *
    * XXX This static variable is not used in a SMP-safe way at first glance.
    *     A deeper look shows that it is OK, because the code is only used on
    *     Linux in the ioctl path, which is protected by the Big Kernel
    *     Lock --hpreg
    */
   static union {
      char f[VM_GET_MEM_INFO_SIZE(MAX_VMS)];
      VMGetMemInfoArgs s;
   } buf;
   VMDriver *vm;

   ASSERT(curVM);

   if (HostIF_CopyFromUser(buf.f, outArgs, VM_GET_MEM_INFO_SIZE(1))) {
      return FALSE;
   }

   HostIF_GlobalVMLock(8);

   /* Now that we have the lock, we can read vmCount --hpreg */
   if (buf.s.numVMs < vmCount) {
      HostIF_GlobalVMUnLock(8);

      return FALSE;
   }

   buf.s.numLockedPages = numLockedPages;
   buf.s.maxLockedPages = lockedPageLimit;

   for (vm = vmDriverList, buf.s.numVMs = 0;
        vm;
        vm = vm->nextDriver, buf.s.numVMs++) {
      ASSERT(buf.s.numVMs < vmCount);
      if (vm == curVM) {
         buf.s.callerIndex = buf.s.numVMs;
      }
      buf.s.memMgmtInfo[buf.s.numVMs] = vm->memMgmtInfo;
   }
   ASSERT(buf.s.numVMs == vmCount);

   HostIF_GlobalVMUnLock(8);

   if (HostIF_CopyToUser(outArgs, buf.f, VM_GET_MEM_INFO_SIZE(buf.s.numVMs))) {
      return FALSE;
   }

   return TRUE;
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_SetMemInfo --
 *
 *      Set the memory management information about this VM.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      vm->memMgmtInfo is filled in.
 *
 *----------------------------------------------------------------------
 */
void
Vmx86_SetMemInfo(VMDriver *curVM,
                 VMSetMemInfoArgs *inArgs)
{
   HostIF_GlobalVMLock(9);

   if (inArgs->info.pageFaultRate >= 0) {
      curVM->memMgmtInfo.pageFaultRate = inArgs->info.pageFaultRate;
   }
   if (inArgs->info.lockFailRate >= 0) {
      curVM->memMgmtInfo.lockFailRate = inArgs->info.lockFailRate;
   }
   if (inArgs->info.maxLocked >= 0) {
      curVM->memMgmtInfo.maxLocked = inArgs->info.maxLocked;
   }
   if (inArgs->info.highWaterMark >= 0) {
      curVM->memMgmtInfo.highWaterMark = inArgs->info.highWaterMark;
   }
   if (inArgs->info.waitingForMemoryTime >= 0) {
      curVM->memMgmtInfo.waitingForMemoryTime = inArgs->info.waitingForMemoryTime;
   }
   if (inArgs->info.poweringOn >= 0) {
      curVM->memMgmtInfo.poweringOn = inArgs->info.poweringOn;
   }
   if (inArgs->info.hasFocus >= 0) {
      curVM->memMgmtInfo.hasFocus = inArgs->info.hasFocus;
   }

   HostIF_GlobalVMUnLock(9);
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_PAEEnabled --
 *
 *      Is PAE enabled?
 *
 * Results:
 *      TRUE if PAE enabled.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
Bool
Vmx86_PAEEnabled()
{
   uint32 cr4;

   GET_CR4(cr4);
   return (cr4 & CR4_PAE) != 0;
}

/*
 *----------------------------------------------------------------------
 *
 * FreeAllResources
 *
 *     Free all the resources allocated for a vm.
 *
 *
 * Results:
 *      None
 *
 * Side effects:
 *      Memory freed.
 *
 *----------------------------------------------------------------------
 */

static void
FreeAllResources(VMDriver *vm)
{
   if (vm) {
#ifdef USE_PERFCTRS_HOSTED
      /*
       * Can't log in PerfCtr_Release() when we're holding the lock
       * and (in any case) the VM has be dequeued.
       */
      PerfCtr_Release(vm, FALSE);
#endif

      HostIF_FreeAllResources(vm);

      HostIF_GlobalVMLock(10);
      numLockedPages -= vm->memMgmtInfo.currentLocked;
      HostIF_GlobalVMUnLock(10);

      if (vm->vmhost != NULL) {
         HostIF_FreeKernelMem(vm->vmhost);
      }
      HostIF_FreeKernelMem(vm);
   }
   VTrace_FreeAllResources();
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_AckUserCall
 *
 *      Acknowledge a user call.  The call need not have completed.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Yes.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_AckUserCall(VMDriver *vm, Vcpuid vcpuid)
{
   VMCrossPage *crosspage = vm->crosspage[vcpuid];

   ASSERT(crosspage->userCallRequest != MODULECALL_USERCALL_NONE);
   crosspage->userCallRequest = MODULECALL_USERCALL_NONE;
   HostIF_AckUserCall(vm, vcpuid);
}


/*
 *----------------------------------------------------------------------
 *
 * Vmx86_CompleteUserCall --
 *
 *      Take actions on completion of a user call, which may or
 *      may not have been acknowledged.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      VCPU thread continues.
 *
 *----------------------------------------------------------------------
 */

void
Vmx86_CompleteUserCall(VMDriver *vm, Vcpuid vcpuid)
{
   VMCrossPage *crosspage = vm->crosspage[vcpuid];

   if (crosspage->userCallRequest != MODULECALL_USERCALL_NONE) {
      Vmx86_AckUserCall(vm, vcpuid);
   }
   HostIF_AwakenVcpu(vm, vcpuid);
}
