/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/


/*
 * modulecall.h
 *
 *        Monitor <-->  Module (kernel driver) interface
 */

#ifndef _MODULECALL_H
#define _MODULECALL_H

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_MONITOR
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#include "includeCheck.h"

#include "x86.h"
#include "iocontrols.h"

/*
 *----------------------------------------------------------------------
 *
 * ModuleCallType --
 *
 *      Enumaration of support calls done by the module
 *
 *----------------------------------------------------------------------
 */

typedef enum ModuleCallType {
   MODULECALL_NONE = 100,
   
   MODULECALL_INTR,                  
   MODULECALL_LOCKPAGE,                  
   MODULECALL_UNLOCKPAGE,                  
   MODULECALL_ISMPNLOCKED, 
   MODULECALL_SEMAWAIT,
   MODULECALL_SEMASIGNAL,
   MODULECALL_SEMAFORCEWAKEUP,
   MODULECALL_IPI,                   // hit thread with IPI

   /*
    * Return codes for user calls
    */

   MODULECALL_USERRETURN,
   MODULECALL_USERTIMEOUT,

   MODULECALL_GET_RECYCLED_PAGE,
   MODULECALL_RELEASE_ANON_PAGE,

   MODULECALL_LAST                   // Number of entries. Must be the last one
} ModuleCallType;

#define MODULECALL_USER_START 300
#define MODULECALL_USER_END   399

#define MODULECALL_CROSS_PAGE_LEN    1
#define MODULECALL_CROSS_PAGE_START  6

// we don't want it in the final version
//#define USE_BLUE_SCREEN

#define MODULECALL_USERCALL_NONE     300

/*
 * Define VMX86_UCCOST in the makefiles (Local.mk,
 * typically) if you want a special build whose only purpose
 * is to measure the overhead of a user call and its
 * breakdown.
 *
 * WINDOWS NOTE: I don't know how to pass VMX86_UCCOST to
 * the driver build on Windows.  It must be defined by hand.
 */

#ifdef VMX86_UCCOST

typedef enum UCCostStamp {
#define UC(x) UCCOST_ ## x,
#include "uccostTable.h"
   UCCOST_MAX
} UCCostStamp;

#define UCTIMESTAMP(cp, stamp) \
             do { (cp)->ucTimeStamps[UCCOST_ ## stamp] = GET_TSC(); } while (0)

#else

#define UCTIMESTAMP(cp, stamp)

#endif

/*
 *----------------------------------------------------------------------
 *
 * VMCrossPage --
 *
 *      data structure shared between the monitor and the module
 *      that is used for crossing between the two.
 *      Accessible as vm->cross (kernel module) and CROSS_PAGE
 *      (monitor)
 *
 *      Exactly one page long
 *
 *----------------------------------------------------------------------
 */

/*
 * We arrived at 1008 as follows. The first 1024 bytes of the cross
 * page contains the version number (4 bytes) and then the task switch
 * code aligned at a 16 byte boundary. This alignment allows the
 * decoders to get a good start. We allocate 1024 bytes for all this
 * although the task switch code is only about 192 bytes large so
 * that the context information (data) that follows this code fragment
 * is on another 1024 byte chunk of memory. This separation makes the
 * trace cache of the P4 happy. If it any closer, the trace cache
 * gets flushed each time we write to the context info data area. 
 */
#define CONTEXT_SWITCH_CODE_SIZE 1008


#define MAX_PASSTHROUGH_PORT 1024
#define NUM_PASSTHROUGH_WORDS (MAX_PASSTHROUGH_PORT / 32)
#define INTERRUPT_REDIRECTION_BITMAP_SIZE 32

#ifdef _WIN32
#define __attribute__(_x)     // NOP
#include <PSHPACK1.H> // 1-byte packing alignment
#endif

typedef struct ContextInfo {
    DTRWords    gdtr;
    DTRWords    ldtr;
    DTRWords    idtr;
    uint16 tr;
    uint8 jump[6];
    Task task;   
} __attribute__ ((packed)) ContextInfo;


typedef struct VMCrossPage {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE];

   /*
    * iobitMap must immediately follow the interrupt redirection
    * bitmap, which must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   struct VMCrossPage *hostVA;

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;

   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;

   MPN crossMonPageTableMPN;

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA crosspageLinearAddr;

   /*
    * The monitor may requests up to two actions when
    * returning to the host.  The moduleCallType field and
    * args encode a request for some action in the driver.
    * The userCallType field (together with the RPC block)
    * encodes a user call request.  The two requests are
    * independent.  The user call is executed first, with
    * the exception of MODULECALL_INTR which has a special
    * effect.
    */
   ModuleCallType moduleCallType;
   uintptr_t args[4];
   uintptr_t retval;

   int userCallType;
   volatile int userCallRequest;	// VCPU/VMX synchronization
   Bool userCallCross;
   Bool userCallRestart;
   
   /*
    * TRUE if moduleCall was interrupted by signal. Only
    * vmmon uses this field to remember that it should
    * restart RunVM call, nobody else should look at it.
    */
   Bool moduleCallInterrupted;
   Bool _pad1[1];

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2];

#ifdef VMX86_UCCOST
    uint64 ucTimeStamps[UCCOST_MAX];
#define UCCOST_CROSSPAGE_SIZE (8 * UCCOST_MAX)
#else
#define UCCOST_CROSSPAGE_SIZE 0
#endif

   /*
    * The values in the shadow debug registers must match
    * those in the hardware debug register immediately after
    * a task switch in either direction.  They are used to
    * minimize moves to and from the debug registers.
    */
   uint32 shadowDR[8];
   uint32 switchPanic;
   Bool monUsesSysenter;

} __attribute__ ((packed)) VMCrossPage;


#ifdef WIN32
#include <POPPACK.H>
#endif


/*
 * We hardcode these constants here instead of using the sizeof() and
 * offsetof() to catch compiler inconsistencies. This file is compiled
 * once onsite to build the monitor, and then once at the customer site
 * to build the drivers.
 *
 * An alternate scheme, and preferred scheme, would be have the makefile
 * generate them, thereby relieving the developer of the error prone
 * chore of updating them each time VMCrossPage is modified. 
 *
 * Pratap 12/01.
 */
#define MODULECALL_CROSSPAGE_SIZE (0x629 + UCCOST_CROSSPAGE_SIZE)

#define MODULECALL_CROSSPAGE_LAOFFSET 0x5d0


#define CROSSPAGE_VERSION (0x177d + VERSION_MAGIC)

#if !defined(VMX86_SERVER) && defined(MONITOR)
#define CROSS_PAGE  ((VMCrossPage *)   \
                     (MONITOR_BASE_VADDR + VPN_2_VA(CROSS_PAGE_START)))
#endif

#define NULLPAGE_LINEAR_START  (MONITOR_LINEAR_START + PAGE_SIZE*NULLPAGE_START)
   

struct passthrough_iorange {
  unsigned short ioBase;	/* Base of range to pass through */
  unsigned short numPorts;	/* Length of range */
};

#define MX_WAITTIMEOUT       100                             // unit: 1ms

#define MX_WAITINTERRUPTED     3
#define MX_WAITTIMEDOUT        2
#define MX_WAITNORMAL          1  // Must equal one; see linux module code.
#define MX_WAITERROR           0  // Use MX_ISWAITERROR() to test for error.

// Any zero or negative value denotes error.
#define MX_ISWAITERROR(e)      ((e) <= MX_WAITERROR)

#endif
